/*
        This program implements (hopefully) the algorithm presented by Chang in

        Numerical Calculation of the inductances of a multi-superconductor 
        transmission line system, IEEE Trans on Mag. MAG-17 No. 1, Jan 1981, p. 764 

        The idea is to minimize the total energy of the system by the variational 
        method.  This entails a minimization function, which is obtained with
        the current distributions over the segments of the conductors.  Using
        Green's function, the magnetic potential is related to the current
        distribution, and the variables result in being only the currents in the
        segments.  With the approximation that the current density is uniform
        within a segement (note: keep the smallest segement <~ penetration
        depth) a closed form can be optained.  
        
        The minimization is then accomplished by using Lagrang multipliers, 
        by setting the derivatives with respect to the current 
        and in the variational parameters to 0.  This results in a set of
        simultanious linear equations.  This is solved, and the current
        distribution is known.  By inversion and summing over the subsegments of
        each seperate inductor, and another matrix inversion, the result is the
        inductance matrix for the conductor system.
*/
/*      I have decided on a data structure and program structure to use.

        The conductors will be read in and stored in a sorted format linked list.
        Each one (group) will then be counted up the total subsegments, and a new linked
        list, one struct per conductor, will be made, with an array of subsegments
        containing lower left, upper right, and 1/area info.  These subsegments were made
        by dividing up the segments in a ratio'ed type as defined in the input:

           type 0 small on ends
           type 1 small on right (upper)
           type 2 small on left (lower)
           type 3 small in middle

        Conductors will be a linked list of stuctures containing the conductor 
        identification number, index start and end, penetration depth (squared),
        and an array of subsegments stuctures: lower left (x,y), upper right (x,y)
        and area.

        Later, I may want to include penetration depth in the subsegment array.

*/

#include "induct.h"

/*=================================================
       main routine
***************************************************/
int
main()
{
 conductor *tmp;
 long m,s;  /* number of elements and conductors */
 long i,j;

  printf("Input is of the form:\n");
  printf("Conductor number, (x,y) lower left, (x,y) upper right,");
  printf("penetraiont depth, x divisions, y divisions, x ratio, y ratio,");
  printf("x type, and y type \n\n");

  printf("The input file is from standard input, so use redirection to use a file.\n");
  printf("Output is to standard output.\n");

/* input file */
  fp = stdin;
/*
  fp = fopen("induct.in","r");
*/

if(fp == NULL)
{
  printf("no input\n");
  return(-1);
}

/* initialize */
  input_cond = NULL;
  cond_head = NULL;
  m = 0;
  s = 0;

/* I will constrain the ground conductors to be numbered >= 100, and that
only these conductors will be used for GND.   getinput() will order it so that
the first conductor, and therefore first element, is GND, and will be the
reference for subsequent P/Q evaluation */
  getinput();

/* debug */
#ifdef DEBUG2
  printf("parsed in input\n");
#endif

/* subdivide and all it needs is in the subseg.c file */
  subdivide();

/* debug */
#ifdef DEBUG
  printf("finished subdivide \n");
#ifdef DEBUG2
  printinputlist();
  printlist();
#endif
#endif
  
  s = 0;
/* the number of elements is equal to the i_end of the last conductor */
  for(tmp=cond_head; tmp->next != NULL; tmp = tmp->next)
           s++;

/* and get the number of elements, this is total, 0 to i_end, so to make a m-1
matrix, just use m.  If there will be a memory problem, this'll be it, Q is huge.
But, since we ignored the first element in subseg, i = -1 to i_end, and the
total segments are now i_end + 2 (1 for 0 and 1 for -1).  Since we want a m-1
matrix, we need i_end + 1 allocated, and will index into it from 0 to i_end */

  m = tmp->i_end + 1;

  q_matrix = (double **) malloc((long)m * sizeof(double *));
  if (!q_matrix) {
    printf("Unable to allocate memory for q matrix\n");
    return(-1);
  }
  for (i=0;i<m;i++) {
    q_matrix[i] = (double *) malloc((long)m * sizeof(double));
    if (!q_matrix[i]) {
      printf("Unable to allocate memory for %dth row in q matrix\n",i);
      return(-1);
    }
  }

  qeval(m);

/* debug */
#ifdef DEBUG
  printf("finished evaluating q matrix\n");
#ifdef DEBUG2
  print_matrix(q_matrix,m,m,"Q matrix");
#endif
#endif

  fast_matrix_invert(q_matrix,m);

/* debug */
#ifdef DEBUG
  printf("inverted q matrix\n");
#ifdef DEBUG2
  print_matrix(q_matrix,m,m,"inverted Q matrix");
#endif
#endif

  s_matrix = (double **) malloc((long)s * sizeof(double *));
  if (!s_matrix) {
    printf("Unable to allocate memory for s matrix\n");
    return(-1);
  }
  for (i=0;i<s;i++) {
    s_matrix[i] = (double *) malloc((long)s * sizeof(double));
    if (!s_matrix[i]) {
      printf("Unable to allocate memory for %dth row in s matrix\n",i);
      return(-1);
    }
  }

  s_qreduce();

/* debug */
#ifdef DEBUG
  printf("reduced q matrix into s matrix \n");
#ifdef DEBUG2
  print_matrix(s_matrix,s,s,"S matrix");
#endif
#endif

  fast_matrix_invert(s_matrix,s);

/* debug */
#ifdef DEBUG
  printf("inverted s matrix \n");
#endif

  print_result(s_matrix,s,s,"***********Inductance matrix***********");

  return (0);
} /* main */

/*=================================================
   reduction of the inverted Q matrix to get s
***************************************************/
/* 
   This routine sums over columns contain i and rows containing j
   to produce a smaller (sxs) matrix where s = # conductors.  With the linked
   lists, this isn't hard.
*/
void
s_qreduce()
{
  long i,j,k,l;
  conductor *tmp1, *tmp2;
  double sum;

  /* don't include GND in the inductance matrix */
  k = 0;
  for(tmp1 = cond_head->next; tmp1 != NULL; tmp1 = tmp1->next)
  {
    l = k;
    for(tmp2 = tmp1; tmp2 != NULL; tmp2 = tmp2->next)
    {
      sum = 0;

/* debug */
/*
printf("s[%d][%d] = sum q[%d][%d] to q[%d][%d]\n",k,l,
          tmp1->i_start,tmp2->i_start,tmp1->i_end,tmp2->i_end);
*/

      for(i=tmp1->i_start; i <= tmp1->i_end; i++)
        for(j=tmp2->i_start; j <= tmp2->i_end; j++)
          sum = sum + q_matrix[i][j];

      s_matrix[k][l] = sum;
      s_matrix[l][k] = sum;
      l++;
    }
    k++;
  }

} /* s_qreduce */


/*=================================================
       evaluation of the elements for the q_matrix
***************************************************/
/*
   Q:mk = 1/2(P:mk + P:km + 2P:MM - P:mM - P:Mm - P:kM - P:Mk) 
   + Uo(L:M^2/A(M) + L:k^2/A(k) d(mk))
   where L is the penetration depth, d(mk) is the Kronecker delta function, A(i)
   is the area of the i'th segment.  NOTE: this was an error in the original
   paper.

   But, we know that P:km = P:mk (by symetry) we get some simplification... 
   Q:mk = P:mk + P:MM - P:mM - P:kM + Uo(L:M^2/A(M) + L:k^2/A(k) d(mk))
   
   Since the inductors are in a linked list, it is faster not to try to 
   "refind" the (xll,yll) ....  to evaluate P, so instead of having a separate
   routine to get Q, I will incorperate it in this routine to calulate all the
   q's for the q_matrix.  It is also important to note that the P:kM only need
   be calculated once/row, this fact will be used to expediate the execution, as
   will that P:MM need be calculated only once.
*/

void
qeval(long m)
{
  long i,j,k;
  double *PM_row,q;
  conductor *tmp1, *tmp2;
  double M_lam_2_area;

  /* first calculate the P:kM (one time only) stored in an array.  Note, this is
  (m+1) long because it will include P:MM.  Although it is a little messy, I
  will use the last element in the conductor as the reference.  Hopefully,
  the rest of the indecies will be consistant in the Q evaluation.*/

  PM_row = (double *) malloc( (long) (m + 1) * sizeof(double));

  tmp1 = cond_head; 
  /* so, P:MM */
  j = tmp1->i_end + 1;
  PM_row[m] = P(tmp1->segarray[j], tmp1->segarray[j]);
/*
printf("P(m,m) = %lf\n",PM_row[m]);
*/

  /* and lambda^2/area for M */
  M_lam_2_area = tmp1->lam2 * tmp1->segarray[j]->area_r;

  /* reset inital i_start to be 0, not -1, to take care of the reference (P:MM) */
  tmp1->i_start = 0;

  /* and now the rest (P:kM) */
  for(tmp2 = tmp1; tmp2 != NULL; tmp2 = tmp2->next)
    for(i = 0; i <= (tmp2->i_end - tmp2->i_start); i++)
{
      PM_row[i + tmp2->i_start] = 
             P(tmp1->segarray[j], tmp2->segarray[i]);
/*
printf("P(m,i) = P(%d,%d) = %lf\n",j,(i+tmp2->i_start),PM_row[i+tmp2->i_start]);
*/
}

  /* now for the q_matrix evaluation... */
  for(tmp1 = cond_head; tmp1 != NULL; tmp1 = tmp1->next)
  {
  /* I will keep i,j for the subseg arrays, and offset them for q_matrix */
    for(i = 0; i <= (tmp1->i_end - tmp1->i_start); i++)
    {
      for(tmp2 = tmp1; tmp2 != NULL; tmp2 = tmp2->next)
      {
      /* since we have linked lists of arrays, either separate loop, or this...  */
        if(tmp2 == tmp1)
          k = i;
        else
          k = 0;

        for(j = k; j <= (tmp2->i_end - tmp2->i_start); j++)
        {
        /* Q:mk = P:mk + P:MM - P:mM - P:kM + Uo(L:M^2/A(M) + L:k^2/A(k) d(mk)) */
          q = P(tmp1->segarray[i],tmp2->segarray[j]) + PM_row[m] 
              - PM_row[tmp1->i_start + i] - PM_row[tmp2->i_start + j];

          if((j == i) && (tmp1 == tmp2))
             q = q + Uo * (M_lam_2_area + 
                 tmp1->lam2 * tmp1->segarray[i]->area_r);
          else
             q = q + Uo * (M_lam_2_area);
             
          /* assign the matrix values, (symetric matrix) */
          q_matrix[i + tmp1->i_start][j + tmp2->i_start] = q;
          q_matrix[j + tmp2->i_start][i + tmp1->i_start] = q;

        } /* inner array loop (j) */
      }   /* inner linked list loop (tmp2) */
    } /* outer array loop (i) */
  }   /* outer linked list loop (tmp1) and  q_matrix evaluation */

} /* qeval */

